package goaml.lang;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import goaml.lang.entity.Entity;
import goaml.lang.process.AbstractProcess;
import goaml.lang.transition.AbstractTransition;
import goaml.lang.util.DelayedUpdate;

public final class StateChart {

	private final String NAME;

	private final List<Entity> ENTITIES;
	private final List<AbstractProcess> PROCESSES;
	private final List<AbstractTransition> TRANSITIONS;
	private final List<DelayedUpdate> DELAYED_UPDATES;

	public StateChart(final String path) {
		this.ENTITIES = new ArrayList<Entity>();
		this.PROCESSES = new ArrayList<AbstractProcess>();
		this.TRANSITIONS = new ArrayList<AbstractTransition>();
		this.DELAYED_UPDATES = new ArrayList<DelayedUpdate>();

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

		factory.setIgnoringComments(true);
		factory.setValidating(true);
		factory.setIgnoringElementContentWhitespace(true);

		try {
			Document xml = factory.newDocumentBuilder().parse(path);

			this.NAME = xml.getDocumentElement().getAttribute("name");

			// Get every entities and parse them
			NodeList entityNodes = xml.getElementsByTagName("Entity");
			for (int i = 0; i < entityNodes.getLength(); i++) {
				this.ENTITIES.addAll(Entity.parse(entityNodes.item(i)));
			}

			// Get every process and parse them
			NodeList processNodes = xml.getElementsByTagName("Process");
			for (int i = 0; i < processNodes.getLength(); i++) {
				AbstractProcess.parse(processNodes.item(i), this);
			}

			// Get every transitions and parse them
			NodeList transitionNodes = xml.getElementsByTagName("Transition");
			for (int i = 0; i < transitionNodes.getLength(); i++) {
				AbstractTransition.parse(transitionNodes.item(i), this);
			}

		} catch (SAXException | IOException | ParserConfigurationException e) {
			e.printStackTrace();
			throw new RuntimeException("Failed to parse the XML file");
		}
	}

	public final void update() {
		// Compute every Process and Transition
		// Process and Transition doesn't change value of the linked AbstractState
		// In this way, all processes and transitions can compute their rate in the same condition
		// and the order of processes and transitions have no consequences
		// Process and Transition instances generate DelayedUpdate when they compute,
		// and these DelayedUpdate can change value of State with the method "apply"
		// which is used after every processes and transitions have compute.
		for (AbstractProcess process : PROCESSES) {
			process.compute(this);
		}

		for (AbstractTransition transition : TRANSITIONS) {
			transition.compute(this);
		}

		// Apply every input and output on corresponding states
		for (DelayedUpdate delayedUpdate : DELAYED_UPDATES) {
			delayedUpdate.apply();
		}

		DELAYED_UPDATES.clear();
	}

	// *************************
	// Getters and setters
	// *************************
	public final String getName() {
		return this.NAME;
	}

	public final List<Entity> getEntitiesByName(final String name) {
		List<Entity> entities = new ArrayList<Entity>();

		for (Entity entity : ENTITIES) {
			if (entity.getName().equals(name)) {
				entities.add(entity);
			}
		}

		return entities;
	}

	public final List<Entity> getEntities() {
		return this.ENTITIES;
	}

	public final void addProcess(final AbstractProcess process) {
		this.PROCESSES.add(process);
	}

	public final void addTransition(final AbstractTransition transition) {
		this.TRANSITIONS.add(transition);
	}

	public final void addDelayedUpdate(final DelayedUpdate delayedUpdate) {
		this.DELAYED_UPDATES.add(delayedUpdate);
	}

}
